import numpy as np

from synthetmic.data.utils import (
    SynthetMicData,
    create_constant_volumes,
    create_periodicity,
    sample_random_seeds,
)


def _create_unit_domain(space_dim: int) -> tuple[np.ndarray, float]:
    match space_dim:
        case 2:
            return np.array([[0, 1], [0, 1]]), 1.0
        case 3:
            return np.array([[0, 1], [0, 1], [0, 1]]), 1.0
        case _:
            raise ValueError(
                f"invalid space_dim: {space_dim}; value must be either 2 or 3."
            )


def create_data_with_constant_volumes(
    space_dim: int = 2,
    n_grains: int = 1000,
    is_periodic: bool = False,
    random_state: int | None = None,
) -> SynthetMicData:
    """
    Create data in a unit square or cube. All target volumes will be the same and seeds will be
    randomly generated in the domain.

    Parameters
    ----------
    space_dim : int, optional
        The spce dimension of the domain. Value must be 2 or 3.

    n_grains : int, optional
        The number of grains in the unit cube. Value must be greater 0.

    is_period : bool, optional
        If True, the corresponding domain will be periodic in all directions.

    random_state : int or None, optional
        Pass an int to make the generated data reproducible.

    Returns
    -------
    synthetmic.data.utils.SyntyhetMicData
    """
    domain, domain_volume = _create_unit_domain(space_dim=space_dim)

    seeds = sample_random_seeds(
        domain=domain, n_grains=n_grains, random_state=random_state
    )
    volumes = create_constant_volumes(n_grains=n_grains, domain_volume=domain_volume)
    periodic = create_periodicity(space_dim=domain.shape[0], is_periodic=is_periodic)

    return SynthetMicData(
        seeds=seeds,
        volumes=volumes,
        domain=domain,
        periodic=periodic,
        init_weights=None,
    )


def create_data_with_lognormal_volumes(
    mean: float = 1.0,
    std: float = 0.35,
    space_dim: int = 2,
    n_grains: int = 1000,
    is_periodic: bool = False,
    random_state: int | None = None,
) -> SynthetMicData:
    """
    Create data in a unit square or cube. All target volumes will be distributed with
    lognormal distribution in the domain.

    Parameters
    ----------
    mean : float, optional
        The mean of the lognormal distribution.

    std : float, optional
        The standard deviation of the lognormal distribution.

    space_dim : int, optional
        The spce dimension of the domain. Value must be 2 or 3.

    n_grains : int, optional
        The number of grains in the unit cube. Value must be greater 0.

    is_period : bool, optional
        If True, the corresponding domain will be periodic in all directions.

    random_state : int or None, optional
        Pass an int to make the generated data reproducible.

    Returns
    -------
    synthetmic.data.utils.SyntyhetMicData
    """
    domain, domain_volume = _create_unit_domain(space_dim=space_dim)

    seeds = sample_random_seeds(
        domain=domain, n_grains=n_grains, random_state=random_state
    )

    ln_std = np.sqrt(np.log(1 + (std / mean) ** 2))
    ln_mean = -0.5 * ln_std**2 + np.log(mean)
    samples = np.random.lognormal(mean=ln_mean, sigma=ln_std, size=n_grains)
    scaling_factor = domain_volume / samples.sum()
    volumes = scaling_factor * samples

    periodic = create_periodicity(space_dim=domain.shape[0], is_periodic=is_periodic)

    return SynthetMicData(
        seeds=seeds,
        volumes=volumes,
        domain=domain,
        periodic=periodic,
        init_weights=None,
    )



import tempfile

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import vtk
from matplotlib import cm, colormaps
from matplotlib.axes import Axes
from vtk.numpy_interface import dataset_adapter as dsa
from vtk.util.numpy_support import vtk_to_numpy

from synthetmic import LaguerreDiagramGenerator


def plot_2dcells_as_matplotlib_fig(
    generator: LaguerreDiagramGenerator,
    ax: Axes | None = None,
    title: str | None = None,
    colorby: np.ndarray | list[float] | None = None,
    colormap: str = "plasma",
    save_path: str | None = None,
) -> Axes:
    """
    A function that plots a 2D Laguerre cells as matplotlib figure.

    Parameters
    ----------

    generator : LaguerreDiagramGenerator
        a fitted LaguerreDiagramGenerator object.
    axis : Axis, optional
        a matplotlib axis object to handle the figure, if None, a new one will be created
    title : str or None, optional
        title of the figure.
    colorby : ndarray or list, shape (N,), optional
        a 1d array of scalars for coloring the cells, if None, the cells will be colored
        by their respective volume.
    colormap: str, optional
        a string representing one of the supported colormaps in the matplotlib library.
    save_path: str or None
        a string reperesenting the path to save the generated figure to, e.g., ./plots/figure2.pdf.
        If None, figure will not be saved.

    Returns
    -------

    ax : matplotlib Axes object
    """
    with tempfile.NamedTemporaryFile(mode="w", suffix=".vtk", delete=True) as tmp_file:
        filename = tmp_file.name

        generator.diagram_to_vtk(filename)

        reader = vtk.vtkUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()

        data = reader.GetOutput()

    N = data.GetNumberOfCells()

    numpy_array_of_cells = dsa.WrapDataObject(data).Cells
    numpy_array_of_points = dsa.WrapDataObject(data).Points

    cell_data = data.GetCellData()
    array = cell_data.GetArray(1)

    # Extract array values
    cell_numbers = []
    for j in range(array.GetNumberOfTuples()):
        tuple_data = array.GetTuple(j)  # Get the data tuple for the cell
        cell_numbers.append(int(tuple_data[0]))

    cells = vtk_to_numpy(numpy_array_of_cells)
    verts = vtk_to_numpy(numpy_array_of_points)

    if ax is None:
        _, ax = plt.subplots()

    if colormap not in list(colormaps):
        raise ValueError(
            f"Invalid colormap string: {colormap}. Value must be one of the following: {', '.join(list(colormaps))}"
        )

    if colorby is None:
        # default to coloring cells by their volumes
        colorby = generator.get_fitted_volumes()

    cmap = cm.get_cmap(colormap)
    norm = mcolors.Normalize(vmin=np.min(colorby), vmax=np.max(colorby))
    colors = cmap(norm(colorby))

    idx = 0
    for k in range(N):
        nv = cells[idx]
        vidx = cells[idx + 1 : idx + nv + 1]
        idx = idx + nv + 1

        ax.fill(
            verts[vidx, 0],
            verts[vidx, 1],
            color=colors[cell_numbers[k]],
            linewidth=1,
            alpha=0.7,
        )

    idx = 0
    for k in range(N):
        nv = cells[idx]
        vidx = cells[idx + 1 : idx + nv + 1]
        idx = idx + nv + 1
        ax.plot(verts[vidx, 0], verts[vidx, 1], "k", linewidth=1)

    if title is not None:
        ax.set_title(title)

    if save_path is not None:
        ax.axis("off")
        ax.set_aspect("equal")

        plt.savefig(save_path, bbox_inches="tight")

    return ax


def plot_cells_as_pyvista_fig(
    generator: LaguerreDiagramGenerator,
    window_size: tuple[int, int] = (1024, 768),
    notebook: bool = False,
    title: str | None = None,
    colorby: np.ndarray | list[float] | None = None,
    colormap: str = "plasma",
    save_path: str | None = None,
    interactive: bool = False,
    include_slices: bool = False,
) -> pv.Plotter:
    """
    A function that plots Laguerre cells (both 2D and 3D) as pyvista figure.

    Parameters
    ----------

    generator : LaguerreDiagramGenerator
        a fitted LaguerreDiagramGenerator object.
    window_size : tuple[int, int], optional
        the figure size.
    notebook : bool, optional
        when True, the resulting plot is placed inline a jupyter notebook. Assumes a jupyter console is active.
        Automatically enables off_screen.
    title : str or None, optional
        title of the figure.
    colorby : ndarray or list, shape (N,), optional
        a 1d array of scalars for coloring the cells, if None, the cells will be colored
        by their respective volume.
    colormap: str, optional
        a string representing one of the supported colormaps in the matplotlib library.
    save_path: str or None
        a string reperesenting the path to save the generated figure to, e.g., ./plots/figure2.pdf.
        If None, figure will not be saved.
    interactive : bool, optional
        when True, figure will be saved as html. This requires you provide save_path and ensure that
        the file name has .html extension.
    include_slices : bool, optional
        when True, include othorgonal slice and slices along the axes coordinates to the figure.

    Returns
    -------

    plotter : pyvista Plotter object
    """
    mesh = generator.get_mesh()

    if colorby is None:
        colorby = generator.get_fitted_volumes()

    # create cell data that gives the cell volumes, this allows us to colour by cell volumes
    mesh.cell_data["vols"] = colorby[mesh.cell_data["num"].astype(int)]

    if colormap not in list(colormaps):
        raise ValueError(
            f"Invalid colormap string: {colormap}. Value must be one of the following: {', '.join(list(colormaps))}"
        )

    if include_slices:
        N_ROW, N_COL = 2, 3
        plotter = pv.Plotter(
            off_screen=True,
            window_size=list(window_size),
            notebook=notebook,
            title=title,
            shape=(N_ROW, N_COL),
        )

        meshes = (
            mesh,
            mesh.slice_orthogonal(),
            mesh.slice_along_axis(n=2, axis="x"),
            mesh.slice_along_axis(axis="x"),
            mesh.slice_along_axis(axis="y"),
            mesh.slice_along_axis(axis="z"),
        )

        c = 0
        for i in range(N_ROW):
            for j in range(N_COL):
                plotter.subplot(i, j)
                plotter.add_mesh(
                    meshes[c],
                    show_edges=True,
                    show_scalar_bar=False,
                )

                c += 1

        plotter.link_views()
        plotter.view_isometric()

    else:
        plotter = pv.Plotter(
            off_screen=True,
            window_size=list(window_size),
            notebook=notebook,
            title=title,
        )
        plotter.add_mesh(
            mesh,
            show_edges=True,
            show_scalar_bar=False,
        )

    if interactive:
        plotter.show_axes()
        plotter.add_scalar_bar(vertical=False)

    if save_path is not None:
        ext = save_path.split(".")[-1]
        if interactive:
            if ext.lower() == "html":
                plotter.export_html(save_path)
            else:
                raise ValueError(
                    f"Invalid file extension: {ext}. Extension must be only html if interactive is set to True."
                )

        else:
            if ext.lower() != "html":
                plotter.save_graphic(save_path)

            else:
                raise ValueError(
                    f"Invalid file extension: {ext}. Extension must not be set to html if interactive is set to False"
                )

    return plotter



from typing import Any, Callable, Type

import numpy as np


class NotFittedError(ValueError, AttributeError):
    """Raised when attempting to use an unfitted generator."""


def _gt(rhs: float) -> Callable[[float | None, str], None]:
    def _out(x: float | None, name: str) -> None:
        if x <= rhs or x is None:
            raise ValueError(f"{name} must be greater than {rhs} but {x} is given.")

        return None

    return _out


def _gte(rhs: float) -> Callable[[float | None, str], None]:
    def _out(x: float | None, name: str) -> None:
        if x < rhs or x is None:
            raise ValueError(
                f"{name} must be greater than or equal to {rhs} but {x} is given."
            )
        return None

    return _out


def _is_instance(
    instance: Type, allow_none: bool = False
) -> Callable[[Any, str], None]:
    def _out(x: Any, name: str) -> None:
        rule = (
            isinstance(x, instance) or (x is None)
            if allow_none
            else isinstance(x, instance)
        )

        if not rule:
            raise TypeError(
                f"{name} must be of type {instance} but {type(x)} is provided."
            )

        return None

    return _out


def _between(
    left: float,
    right: float,
    left_open: bool = False,
    right_open: bool = False,
    both_open: bool = False,
) -> Callable[[float | None, str], None]:
    def _rule(x: float) -> bool:
        if left_open:
            return left < x <= right

        if right_open:
            return left <= x < right

        if both_open:
            return left < x < right

        return left <= x <= right

    def _out(x: float | None, name: str) -> None:
        if (not _rule(x)) or x is None:
            raise ValueError(
                f"{name} must be between {left} and {right}, but {x} is given."
            )

        return None

    return _out


def _compose_rules(*args) -> Callable:
    rule_fns = [arg for arg in args if callable(arg)]

    def _out(x: Any, name: str):
        for rule_fn in rule_fns:
            res = rule_fn(x, name)
            if res is not None:
                return res

    return _out


def _check_array(
    allowed_types: list[Type], allowed_shapes: list[tuple[int, int]] | None = None
) -> Callable[[np.ndarray, str], None]:
    def _out(x: np.ndarray, name: str) -> None:
        if x.size == 0:
            raise ValueError(f"{name} is empty. Input required a non-empty ndarray.")

        if x.dtype not in allowed_types:
            raise ValueError(
                f"{name} contain elements of wrong type {x.dtype}. Allowed types are {allowed_types}."
            )

        if allowed_shapes is not None:
            if x.shape not in allowed_shapes:
                raise ValueError(
                    f"{name} hase a wrong shape {x.shape}. Allowed shapes are {allowed_shapes}."
                )

        return None

    return _out


def _check_periodic(x: list[bool], name: str) -> None:
    if len(x) not in (2, 3):
        raise ValueError(
            f"invalid {name} length {len(x)}; expected length to be 2 or 3."
        )

    if not all(isinstance(var, bool) for var in x):
        raise ValueError(f"all entries in {name} must be bool.")

    return None


def validate_generator_params(
    tol: float,
    n_iter: int,
    damp_param: float,
    verbose: bool,
) -> None:
    _compose_rules(_is_instance(instance=float), _gt(rhs=0.0))(tol, "tol")
    _compose_rules(_is_instance(instance=int), _gte(rhs=0))(n_iter, "n_iter")
    _compose_rules(_is_instance(instance=float), _between(left=0.0, right=1.0))(
        damp_param, "damp_param"
    )
    _is_instance(instance=bool)(verbose, "verbose")

    return None


def validate_fit_args(
    seeds: np.ndarray,
    volumes: np.ndarray,
    domain: np.ndarray,
    periodic: list[bool] | None,
    init_weights: np.ndarray | None,
) -> None:
    _compose_rules(
        _is_instance(instance=np.ndarray), _check_array(allowed_types=[float, int])
    )(seeds, "seeds")

    _compose_rules(
        _is_instance(instance=np.ndarray), _check_array(allowed_types=[float, int])
    )(volumes, "volumes")

    _compose_rules(
        _is_instance(instance=np.ndarray),
        _check_array(allowed_types=[float, int], allowed_shapes=[(2, 2), (3, 2)]),
    )(domain, "domain")

    _is_instance(instance=list, allow_none=True)(periodic, "periodic")
    if periodic is not None:
        _check_periodic(periodic, "periodic")

    _is_instance(instance=np.ndarray, allow_none=True)(init_weights, "init_weights")
    if init_weights is not None:
        _check_array(allowed_types=[float, int])(init_weights, "init_weights")

    # check if the number of samples match
    num_samples = []
    for x in (seeds, volumes, init_weights):
        if x is not None:
            num_samples.append(x.shape[0])

    if len(set(num_samples)) > 1:
        raise ValueError(
            f"one or more of seeds, volumes, and init_weights have inconsistent number of samples: {num_samples}."
        )

    # check if space dimensions match
    space_dims = [seeds.shape[1], domain.shape[0]]
    if periodic is not None:
        space_dims.append(len(periodic))

    if len(set(space_dims)) > 1:
        raise ValueError(
            f"one or more of seeds, domain, and periodic have inconsistent space dimension: {space_dims}."
        )

    if not set(space_dims).issubset({2, 3}):
        raise ValueError(f"""one or more of seeds, domain, and periodic have wrong space dimension: {space_dims}.
            Supported space dimensions are 2 and 3.""")

    return None



from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class SynthetMicData:
    seeds: np.ndarray
    volumes: np.ndarray
    domain: np.ndarray
    periodic: list[bool] | None
    init_weights: np.ndarray | None


def sample_random_seeds(
    domain: np.ndarray, n_grains: int, random_state: int | None = None
) -> np.ndarray:
    np.random.seed(seed=random_state)

    return np.column_stack(
        [np.random.uniform(low=d[0], high=d[1], size=n_grains) for d in domain]
    )


def create_periodicity(space_dim: int, is_periodic: bool) -> list[bool] | None:
    return [True] * space_dim if is_periodic else None


def create_constant_volumes(
    n_grains: int,
    domain_volume: float,
) -> np.ndarray:
    return (np.ones(n_grains) / n_grains) * domain_volume



import math
from enum import StrEnum, auto

import numpy as np

from synthetmic.data.utils import (
    SynthetMicData,
    create_periodicity,
    sample_random_seeds,
)


class _Initializer(StrEnum):
    RANDOM = auto()
    BANDED = auto()
    CLUSTERED = auto()
    MIXED_BANDED_AND_RANDOM = auto()


class _Gradient(StrEnum):
    INCREASING = auto()
    LARGE_AT_MIDDLE = auto()


def _calulate_rel_vols(n1: int, n2: int, r: int) -> np.ndarray:
    vols = np.concatenate((np.ones(n1), r * np.ones(n2)))

    return vols / np.sum(vols)


def _create_layered_points(
    n_layer_arr: np.ndarray, r_layer_arr: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    if isinstance(n_layer_arr, list):
        n_layer_arr = np.array(n_layer_arr)

    if isinstance(n_layer_arr, list):
        r_layer_arr = np.array(r_layer_arr)

    # Relative volumes of the grains (relative to the volume of the box)
    rel_vols = np.repeat(r_layer_arr, n_layer_arr)
    total_vol: float = (n_layer_arr * r_layer_arr).sum()
    rel_vols = rel_vols / total_vol

    # Thickness of each layer
    t_layer_arr = n_layer_arr * r_layer_arr / total_vol
    o_layer_arr = t_layer_arr.cumsum()
    o_layer_arr = np.hstack(([0], o_layer_arr[:-1]))

    # Deduce the z-coord
    z_coord = []
    for i, n in enumerate(n_layer_arr):
        z_coord.extend((o_layer_arr[i] + t_layer_arr[i] * np.random.rand(n)).tolist())

    DIM = 3
    z_coord = np.array(z_coord)
    xy_coord = np.random.rand(n_layer_arr.sum(), DIM - 1)

    return np.column_stack((xy_coord, z_coord)), rel_vols


def _create_subdomains(domain: np.ndarray, n_subdomains: int) -> list[np.ndarray]:
    x_min, x_max = domain[0]
    y_min, y_max = domain[1]

    x_divisions = np.linspace(x_min, x_max, n_subdomains + 1)

    subdomains = []
    for i in range(n_subdomains):
        subdomain = np.array([[x_divisions[i], x_divisions[i + 1]], [y_min, y_max]])
        subdomains.append(subdomain)

    return subdomains


def _create_banded_points(
    subdomains: list[np.ndarray],
    n_points_small: int,
    n_points_large: int,
    volume_frac: float,
) -> tuple[list, list]:
    X = []
    y = []
    for i, subdomain in enumerate(subdomains):
        if i % 2 == 0:
            X.extend(sample_random_seeds(subdomain, n_points_large).tolist())
            y.extend([20 * volume_frac] * n_points_large)
        else:
            X.extend(sample_random_seeds(subdomain, n_points_small).tolist())
            y.extend([volume_frac] * n_points_small)

    return X, y


def _create_uniform_disc_points(
    center: tuple[float, float], radius: float, n_points: int
) -> np.ndarray:
    r = radius * np.sqrt(np.random.random(n_points))
    theta = 2 * np.pi * np.random.random(n_points)
    x = center[0] + r * np.cos(theta)
    y = center[1] + r * np.sin(theta)

    return np.column_stack([x, y])


def _generate_discs(
    centers: list[tuple[float, float]], radius: float, n_points: int
) -> list[np.ndarray]:
    all_points = []

    for center in centers:
        points = _create_uniform_disc_points(center, radius, n_points)
        all_points.append(points)

    return all_points


def _sample_points_outside_discs(
    rect: np.ndarray,
    disc_centers: list[tuple[float, float]],
    disc_radius: float,
    n_points: int,
) -> np.ndarray:
    accepted = []
    batch_size = max(1000, n_points * 2)
    radius_sq = disc_radius**2
    centers_array = np.array(disc_centers)

    while len(accepted) < n_points:
        points = sample_random_seeds(rect, batch_size)

        dists_sq = np.sum(
            (points[:, np.newaxis, :] - centers_array[np.newaxis, :, :]) ** 2, axis=2
        )

        outside_mask = np.all(dists_sq > radius_sq, axis=1)

        valid_points = points[outside_mask]
        accepted.extend(valid_points.tolist())

    accepted = np.array(accepted[:n_points])
    return accepted


def create_example3_data(is_periodic: bool) -> SynthetMicData:
    """
    Create data for generating figures 1 and 2 of the following paper:

    Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
    Laguerre tessellations and polycrystalline microstructures:
    A fast algorithm for generating grains of given volumes,
    Philosophical Magazine, 100, 2677-2707. Link:
    https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

    Parameters
    ----------
    is_periodic : bool
        If True, the underlying domain will be periodic in all
        directions.

    Returns
    -------
    synthetmic.data.utils.SynthetMicData
    """

    domain = np.array([[0, 1], [0, 1]])

    N_GRAINS = 50
    AREA_FRAC = 1 / 185

    y = np.zeros(N_GRAINS)
    y[:35] = AREA_FRAC
    y[35:] = 10 * AREA_FRAC

    X = sample_random_seeds(domain, N_GRAINS)

    periodic = create_periodicity(domain.shape[0], is_periodic)

    return SynthetMicData(
        seeds=X, volumes=y, domain=domain, periodic=periodic, init_weights=None
    )


def create_example4_data(initializer: str, is_periodic: bool) -> SynthetMicData:
    """
    Create data for generating figure 4 of the following paper:

    Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
    Laguerre tessellations and polycrystalline microstructures:
    A fast algorithm for generating grains of given volumes,
    Philosophical Magazine, 100, 2677-2707. Link:
    https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

    Parameters
    ----------
    initializer : str
        The initializer for the generated data. Value must be one of
        ["random", "banded", "clustered", "mixed_banded_and_random"].

    is_periodic : bool
        If True, the underlying domain will be periodic in all
        directions.

    Returns
    -------
    synthetmic.data.utils.SynthetMicData
    """

    AREA_FRAC = 1 / 800
    N1 = 800
    N2 = 200

    domain = np.array([[0, 3], [0, 2]])

    match initializer:
        case _Initializer.RANDOM:
            y = np.zeros(N1 + N2)
            y[:N1] = AREA_FRAC
            y[N1:] = 20 * AREA_FRAC

            X = sample_random_seeds(domain, N1 + N2)

        case _Initializer.BANDED:
            subdomains = _create_subdomains(domain=domain, n_subdomains=7)

            X, y = _create_banded_points(
                subdomains=subdomains,
                n_points_small=math.ceil(N1 / 3),
                n_points_large=math.ceil(N2 / 4),
                volume_frac=AREA_FRAC,
            )
            X = np.array(X)[: N1 + N2]
            y = np.array(y)[: N1 + N2]

        case _Initializer.CLUSTERED:
            X = []
            y = []

            DISC_CENTERS = [(0.6, 0.6), (2.4, 0.6), (1.5, 1.5)]
            DISC_RADIUS = 0.2

            in_disc_points = _generate_discs(
                centers=DISC_CENTERS,
                radius=DISC_RADIUS,
                n_points=math.ceil(N1 / 3),
            )
            for point in in_disc_points:
                X.extend(point.tolist())
                y.extend([AREA_FRAC] * point.shape[0])

            # remove the last point; we have more than N1 points now
            X = X[:-1]
            y = y[:-1]

            # generate points outside circular discs
            out_disc_points = _sample_points_outside_discs(
                rect=domain,
                disc_centers=DISC_CENTERS,
                disc_radius=DISC_RADIUS,
                n_points=N2,
            )
            X.extend(out_disc_points.tolist())
            y.extend([20 * AREA_FRAC] * out_disc_points.shape[0])

            X = np.array(X)
            y = np.array(y)

        case _Initializer.MIXED_BANDED_AND_RANDOM:
            subdomains = _create_subdomains(domain=domain, n_subdomains=7)

            n_points_small = math.ceil(N1 / 4)
            X, y = _create_banded_points(
                subdomains=subdomains,
                n_points_small=n_points_small,
                n_points_large=math.ceil(N2 / 4),
                volume_frac=AREA_FRAC,
            )

            # add random N2 / 4 random points for the small cells
            X.extend(sample_random_seeds(domain, n_points_small).tolist())
            y.extend([AREA_FRAC] * n_points_small)

            X = np.array(X)
            y = np.array(y)

        case _:
            raise ValueError(
                f"initializer must be one of {', '.join(_Initializer)}; but {initializer} was given."
            )

    periodic = create_periodicity(domain.shape[0], is_periodic)

    return SynthetMicData(
        seeds=X, volumes=y, domain=domain, periodic=periodic, init_weights=None
    )


def create_example4b_data(gradient: str, is_periodic: bool) -> SynthetMicData:
    """
    Create data for generating figure 5 of the following paper:

    Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
    Laguerre tessellations and polycrystalline microstructures:
    A fast algorithm for generating grains of given volumes,
    Philosophical Magazine, 100, 2677-2707. Link:
    https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

    Parameters
    ----------
    gradient: str
        The gradient type for the generated data. Value must be one of
        ["increasing", "large_at_middle"].

    is_periodic : bool
        If True, the underlying domain will be periodic in all
        directions.

    Returns
    -------
    synthetmic.data.utils.SynthetMicData
    """

    if gradient not in _Gradient:
        raise ValueError(
            f"gradient must be one of {', '.join(_Gradient)}; but {gradient} was given."
        )

    N = 1000
    MAX_RATIO = 100

    domain = np.array([[0, 3], [0, 2]])
    total_area = (domain[0][1] - domain[0][0]) * (domain[1][1] - domain[1][0])

    uniform_areas = 1 + np.random.uniform(0, 1, N) * (MAX_RATIO - 1)
    # scale so total area equals target
    scaling_factor = total_area / np.sum(uniform_areas)
    y = scaling_factor * uniform_areas

    y = np.sort(y)

    x_coord = np.linspace(start=domain[0][0], stop=domain[0][1], num=N)
    y_coord = np.random.uniform(low=domain[1][0], high=domain[1][1], size=N)
    X = np.column_stack((x_coord, y_coord))

    periodic = create_periodicity(domain.shape[0], is_periodic)

    if gradient == _Gradient.INCREASING:
        return SynthetMicData(
            seeds=X, volumes=y, domain=domain, periodic=periodic, init_weights=None
        )

    y = np.concatenate(
        (y[len(y) % 2 :: 2], y[::-2])
    )  # ensure the large areas fall in the middle and small areas fall at the ends

    return SynthetMicData(
        seeds=X, volumes=y, domain=domain, periodic=periodic, init_weights=None
    )


def create_example5p1_data(n_grains: int, r: int, is_periodic: bool) -> SynthetMicData:
    """
    Create data for generating figure 6 of the following paper:

    Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
    Laguerre tessellations and polycrystalline microstructures:
    A fast algorithm for generating grains of given volumes,
    Philosophical Magazine, 100, 2677-2707. Link:
    https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

    Parameters
    ----------
    n_grains : int
        The number of grains in the microstructure.

    r : int
        The volume ratio of the idealized dual-phase
        microstructure. If the total volume of the first phase
        is v, then that of the second phase will be r*v.

    is_periodic : bool
        If True, the underlying domain will be periodic in all
        directions.

    Returns
    -------
    synthetmic.data.utils.SynthetMicData
    """

    L1, L2, L3 = 100, 100, 100

    domain = np.array([[0, L1], [0, L2], [0, L3]])
    domain_vol = np.prod(domain[:, 1] - domain[:, 0])

    periodic = create_periodicity(domain.shape[0], is_periodic)

    X = sample_random_seeds(domain, n_grains)
    target_vols = domain_vol * _calulate_rel_vols(n_grains // 2, n_grains // 2, r)

    return SynthetMicData(
        seeds=X,
        volumes=target_vols,
        domain=domain,
        periodic=periodic,
        init_weights=None,
    )


def create_example5p4_data(is_periodic: bool) -> SynthetMicData:
    """
    Create data for generating figure 12 of the following paper:

    Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
    Laguerre tessellations and polycrystalline microstructures:
    A fast algorithm for generating grains of given volumes,
    Philosophical Magazine, 100, 2677-2707. Link:
    https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

    Parameters
    ----------
    is_periodic : bool
        If True, the underlying domain will be periodic in all
        directions.

    Returns
    -------
    synthetmic.data.utils.SynthetMicData
    """

    L1 = 2
    L2 = 3
    L3 = 2

    domain = np.array([[0, L1], [0, L2], [0, L3]])
    domain_vol = np.prod(domain[:, 1] - domain[:, 0])

    periodic = create_periodicity(domain.shape[0], is_periodic)

    # Number of grains and relative vols
    n_layer_arr = np.array([1000, 8000, 1000])
    r_layer_arr = np.array([1, 0.05, 1])

    # Target volumes and initial seed locations
    X, rel_vols = _create_layered_points(n_layer_arr, r_layer_arr)
    X = X @ np.diag([L1, L2, L3])
    y = rel_vols * domain_vol

    return SynthetMicData(
        seeds=X, volumes=y, domain=domain, periodic=periodic, init_weights=None
    )


def create_example5p5_data(is_periodic: bool) -> SynthetMicData:
    """
    Create data for generating figure 13 of the following paper:

    Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
    Laguerre tessellations and polycrystalline microstructures:
    A fast algorithm for generating grains of given volumes,
    Philosophical Magazine, 100, 2677-2707. Link:
    https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

    Parameters
    ----------
    is_periodic : bool
        If True, the underlying domain will be periodic in all
        directions.

    Returns
    -------
    synthetmic.data.utils.SynthetMicData
    """

    L1 = 2
    L2 = 2
    L3 = 2

    domain = np.array([[0, L1], [0, L2], [0, L3]])
    domain_vol = np.prod(domain[:, 1] - domain[:, 0])

    # Set periodicity in the three directions
    periodic = create_periodicity(domain.shape[0], is_periodic)

    # Number of grains
    N = 10000

    # Random initial seed locations
    X = sample_random_seeds(domain, N)

    # Log-normal target volumes
    DESIRED_LN_MEAN = 1
    DESIRED_LN_STD = 0.35

    sigma = np.sqrt(np.log(1 + (DESIRED_LN_STD / DESIRED_LN_MEAN) ** 2))
    mu = -0.5 * sigma**2 + np.log(DESIRED_LN_MEAN)

    radii = np.random.lognormal(mu, sigma, N)
    target_vols = radii**3
    rel_vols = target_vols / np.sum(target_vols)

    target_vols = rel_vols * domain_vol

    return SynthetMicData(
        seeds=X,
        volumes=target_vols,
        domain=domain,
        periodic=periodic,
        init_weights=None,
    )



import itertools
import tempfile
from pathlib import Path
from typing import Any, Self

import numpy as np
import pyvista as pv
from pysdot import OptimalTransport
from pysdot.domain_types import ConvexPolyhedraAssembly

from synthetmic.utils import (
    NotFittedError,
    validate_fit_args,
    validate_generator_params,
)


class LaguerreDiagramGenerator:
    """
    tol : float, optional
        Relative percentage error for volumes.
    n_iter : int, optional
        Number of iterations of Lloyd's algorithm (move each seed to the
        centroid of its cell). If it is set to 0, then no Lloyd's iteration
        will be performed.
    damp_param : float [0, 1], optional
        The damping parametr of the damped Lloyd step; value must be between
        0 and 1 (inclusive at both ends).
    verbose : bool, optional
        If set to True, print optimisation progress.
    """

    def __init__(
        self,
        tol: float = 1.0,
        n_iter: int = 5,
        damp_param: float = 1.0,
        verbose: bool = True,
    ):
        self.tol = tol
        self.n_iter = n_iter
        self.damp_param = damp_param
        self.verbose = verbose

        self.space_dim_: int | None = None
        self.optimal_transport_: OptimalTransport | None = None
        self.max_percentage_error_: float | None = None
        self.mean_percentage_error_: float | None = None

    def _print_msg(self, msg: str) -> None:
        if self.verbose:
            print(msg)

    def _update_optimal_transport(self, optimal_transport: OptimalTransport) -> None:
        self.optimal_transport_ = optimal_transport

    def _update_errors(self, y: np.ndarray) -> None:
        percentage_errors = np.array(
            100.0 * np.abs(self.optimal_transport_.pd.integrals() - y) / y
        )
        self.max_percentage_error_ = percentage_errors.max()
        self.mean_percentage_error_ = percentage_errors.mean()

    def _ensure_fitted(self) -> NotFittedError | None:
        """Ensure generator is fitted before use."""

        if self.optimal_transport_ is None:
            raise NotFittedError(
                f"This {self.__class__.__name__} instance is not fitted yet. "
                f"Call 'fit' with appropriate arguments before using this estimator."
            )

        return None

    def fit(
        self,
        seeds: np.ndarray,
        volumes: np.ndarray,
        domain: np.ndarray,
        periodic: list[bool] | None = None,
        init_weights: np.ndarray | None = None,
    ) -> Self:
        """This function implements Algorithm 1 and 2 from the following paper:

        Bourne, D.P., Kok, P.J.J., Roper, S.M. & Spanjer, W.D.T. (2020)
        Laguerre tessellations and polycrystalline microstructures:
        A fast algorithm for generating grains of given volumes,
        Philosophical Magazine, 100, 2677-2707.
        https://www.tandfonline.com/doi/full/10.1080/14786435.2020.1790053

        Parameters
        ----------

        seeds : ndarray, shape (N,d)
            Locations of the N seeds.
        volumes : ndarray, shape (N,)
            Target volumes or areas of the N Laguerre cells.
        domain : ndarray, shape (d,2)
            minimum and maximum coordinates of the box in each of the d dimensions
            (d=2,3)
        periodic : list, optional, length d
            List of Booleans indicating whether or not the domain is periodic in
            the different directions. None indicates no periodicity in any direction.
        init_weights : ndarray, optional, shape (N,)
            Initial guess for the weights for Algorithm 1.
            None indicates that the weights should be zero.

        Returns
        -------

        self

        """

        validate_generator_params(
            tol=self.tol,
            n_iter=self.n_iter,
            damp_param=self.damp_param,
            verbose=self.verbose,
        )

        validate_fit_args(
            seeds=seeds,
            volumes=volumes,
            domain=domain,
            periodic=periodic,
            init_weights=init_weights,
        )

        self.space_dim_ = domain.shape[0]

        # Build the domain
        omega = ConvexPolyhedraAssembly()

        mins = domain[:, 0].copy()
        maxs = domain[:, 1].copy()
        lens = domain[:, 1] - domain[:, 0]
        if periodic is not None:
            for k, p in enumerate(periodic):
                if p:
                    mins[k] = mins[k] - lens[k]
                    maxs[k] = maxs[k] + lens[k]
        omega.add_box(mins, maxs)

        # Turn the relative percentage error into an absolute error tolerance
        # by using the smallest volume
        err_tol = np.min(volumes) * self.tol / 100.0

        # If weight is None, then set the initial guess to zero
        if init_weights is None:
            init_weights = np.zeros(seeds.shape[0])

        # Set up the optimal transport problem
        optimal_transport = OptimalTransport(
            positions=seeds,
            masses=volumes,
            weights=init_weights,
            domain=omega,
            obj_max_dm=err_tol,
            verbosity=0,
        )

        # If there is periodicity, then add the replicants
        if periodic is not None:
            periodic_dict = {True: [-1, 0, 1], False: [0]}
            periodic_list = [periodic_dict[p] for p in periodic]

            cartesian_periodic = list(itertools.product(*periodic_list))

            for rep in cartesian_periodic:
                if rep != (0, 0, 0):
                    optimal_transport.pd.add_replication(rep * lens)

        if self.n_iter == 0:
            # Algorithm 1
            # Solve the optimal transport problem (solve for the weights)
            optimal_transport.adjust_weights()

            self._update_optimal_transport(optimal_transport=optimal_transport)
            self._update_errors(y=volumes)

            return self

        # Algorithm 2
        MIN_VOL_TOL = 1e-10
        x_0 = seeds
        for k in range(self.n_iter):
            # Calculate the centroids: damped Lloyd step
            x_k = (
                1 - self.damp_param
            ) * x_0 + self.damp_param * optimal_transport.get_centroids()

            # Update the seed positions
            optimal_transport.set_positions(x_k)

            # Solve the optimal transport problem.
            # If moving the seeds and maintaining the weights gives a cell that is
            # empty, then we have a bad initial guess for the OT solver.
            # Check whether the smallest volume is bigger than some tolerance:
            # if so, then use the same weights; if not, reset the weights to zero.
            m = optimal_transport.pd.integrals()
            if np.min(m) > MIN_VOL_TOL:
                # Solve for the weights
                optimal_transport.adjust_weights()
            else:
                self._print_msg("Resetting weights to init_weights")
                optimal_transport.set_weights(init_weights)
                # Solve for the weights
                optimal_transport.adjust_weights()

            self._update_optimal_transport(optimal_transport=optimal_transport)
            self._update_errors(y=volumes)

            self._print_msg(
                f"iteration: {k + 1}/{self.n_iter}, max_percentage_error: {self.max_percentage_error_:.4f}%, "
                f"mean_percentage_error: {self.mean_percentage_error_:.4f}%"
            )

            x_0 = x_k

        return self

    def get_fitted_volumes(self) -> np.ndarray:
        """Get the fitted volumes after fitting generator."""
        self._ensure_fitted()

        return self.optimal_transport_.pd.integrals()

    def get_mesh(self) -> pv.PolyData | pv.UnstructuredGrid:
        """Get the underlying mesh as a pyvista PolyData or UnstructuredGrid data object."""

        self._ensure_fitted()

        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".vtk", delete=True
        ) as tmp_file:
            filename = tmp_file.name

            self.optimal_transport_.pd.display_vtk(
                filename, points=None, centroids=None
            )

            mesh = pv.read(filename)

        return mesh

    def get_positions(self) -> np.ndarray:
        """Get the final positions of initial seeds used for generating the laguerre diagram."""

        self._ensure_fitted()

        return self.optimal_transport_.pd.get_positions()

    def get_centroids(self) -> np.ndarray:
        """Get the centroids of the cells in the laguerre diagram."""

        self._ensure_fitted()

        return self.optimal_transport_.pd.centroids()

    def get_vertices(self) -> dict[int, list]:
        """Get the vertices of the cells in the laguerre diagram.

        Return
        ------
        A dictionary with keys as cell ids and values as the
        corresponding vertices.

        In 2D, the format looks like this:

        {
            0: [v_1, v_2, ...],
            ...
            n-1: [v_1, v_2, ...],

        }
        where n is the number of cells or grains.

        In 3D, the format looks like this:

        {
            0: [[v_1, v_2, ...], [v_1, v_2, ...], ...],
            ...
            n-1: [[v_1, v_2, ...], [v_1, v_2, ...], ...],

        }
        where n is the number of cells or grains. Note that the inner
        list of vertices for each cell corresponds to the face vertices.
        """

        self._ensure_fitted()

        res = {}

        if self.space_dim_ == 2:
            offsets, coords = self.optimal_transport_.pd.cell_polyhedra()

            for i in range(len(offsets) - 1):
                s, e = offsets[i : i + 2]
                res[i] = coords[s:e].tolist()

        elif self.space_dim_ == 3:
            offsets_polyhedra, offsets_polygon, coords = (
                self.optimal_transport_.pd.cell_polyhedra()
            )

            for i in range(len(offsets_polyhedra) - 1):
                s1, e1 = offsets_polyhedra[i : i + 2]
                sub_offsets = offsets_polygon[
                    s1 : e1 + 1
                ]  # 1 is added to e1 to get the correct vertex index

                cell_vertices = []

                for j in range(len(sub_offsets) - 1):
                    s2, e2 = sub_offsets[j : j + 2]

                    cell_vertices.append(coords[s2:e2].tolist())

                res[i] = cell_vertices

        return res

    def get_weights(self) -> np.ndarray:
        """Get the weights of the laguerre diagram."""

        self._ensure_fitted()

        return self.optimal_transport_.pd.get_weights()

    def diagram_to_vtk(self, filename: str | Path) -> None:
        """Write the generated diagram to .vkt file; filename must ends with .vtk."""

        self._ensure_fitted()

        self.optimal_transport_.pd.display_vtk(
            filename=filename, points=False, centroids=False
        )

        return None

    def get_params(self) -> dict[str, Any]:
        """Get the parameters of this instance as a dictionary."""

        return dict(
            tol=self.tol,
            n_iter=self.n_iter,
            damp_param=self.damp_param,
            verbose=self.verbose,
        )
